package meerkat.voting.controller;

import com.google.protobuf.ByteString;
import meerkat.protobuf.Voting.*;
import meerkat.voting.controller.callbacks.*;
import meerkat.voting.controller.commands.*;
import meerkat.voting.controller.selector.QuestionSelector;
import meerkat.voting.encryptor.VotingBoothEncryptor;
import meerkat.voting.output.BallotOutputDevice;
import meerkat.voting.storage.StorageManager;
import meerkat.voting.ui.VotingBoothUI;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.concurrent.LinkedBlockingQueue;

/**
 * Created by hai on 28/03/16.
 */
public class VotingBoothImpl implements VotingBoothController, Runnable {

    private BallotOutputDevice outputDevice;
    private VotingBoothEncryptor encryptor;
    private VotingBoothUI ui;
    private StorageManager storageManager;
    private BallotQuestion[] questionsForChoosingChannel;
    private BallotQuestion[] questions;
    private QuestionSelector questionSelector;

    private LinkedBlockingQueue<ControllerCommand> queue;

    private Logger logger;

    private ControllerState state;
    private boolean shutDownHasBeenCalled;

    protected final int MAX_REQUEST_IDENTIFIER = 100000;
    private static int requestCounter = 0;

    private UIElement waitForCommitMessage;
    private UIElement waitForAuditMessage;
    private UIElement waitForCastMessage;



    public VotingBoothImpl () {
        logger = LoggerFactory.getLogger(VotingBoothImpl.class);
        logger.info("A VotingBoothImpl is constructed");
        shutDownHasBeenCalled = false;
        queue = new LinkedBlockingQueue<>();

        waitForCommitMessage = UIElement.newBuilder()
                .setType(UIElementDataType.TEXT)
                .setData(ByteString.copyFromUtf8("Please wait while committing to ballot"))
                .build();
        waitForAuditMessage = UIElement.newBuilder()
                .setType(UIElementDataType.TEXT)
                .setData(ByteString.copyFromUtf8("Please wait while auditing your ballot"))
                .build();
        waitForCastMessage = UIElement.newBuilder()
                .setType(UIElementDataType.TEXT)
                .setData(ByteString.copyFromUtf8("Please wait while finalizing your ballot for voting"))
                .build();

        state = new VotingBoothImpl.ControllerState();

    }

    @Override
    public void init(BallotOutputDevice outputDevice,
                     VotingBoothEncryptor vbEncryptor,
                     VotingBoothUI vbUI,
                     StorageManager vbStorageManager) {
        logger.info("init is called");
        this.outputDevice = outputDevice;
        this.encryptor = vbEncryptor;
        this.ui = vbUI;
        this.storageManager = vbStorageManager;
    }

    @Override
    public void setBallotChannelChoiceQuestions(BallotQuestion[] questions) {
        logger.info("setting questions");
        this.questionsForChoosingChannel = questions;
    }

    @Override
    public void setBallotChannelChoiceQuestions(ArrayList<BallotQuestion> questions) {
        BallotQuestion[] bqArr = new BallotQuestion[questions.size()];
        for (int i = 0; i < bqArr.length; ++i) {
            bqArr[i] = questions.get(i);
        }
        setBallotChannelChoiceQuestions(bqArr);
    }

    @Override
    public void setChannelQuestionSelector(QuestionSelector selector) {
        this.questionSelector = selector;
    }

    @Override
    public void setBallotRaceQuestions(BallotQuestion[] questions) {
        logger.info("setting questions");
        this.questions = questions;
    }

    @Override
    public void setBallotRaceQuestions(ArrayList<BallotQuestion> questions) {
        // TODO: why does the toArray method not work??
        //BallotQuestion[] bqArr = (BallotQuestion[])(questions.toArray());
        BallotQuestion[] bqArr = new BallotQuestion[questions.size()];
        for (int i = 0; i < bqArr.length; ++i) {
            bqArr[i] = questions.get(i);
        }
        setBallotRaceQuestions(bqArr);
    }

    @Override
    public void run() {
        logger.info("run command has been called");
        runVotingFlow();
        doShutDown();

    }

    @Override
    public void shutDown() {
        logger.info("shutDown command has been called");
        synchronized (this) {
            shutDownHasBeenCalled = true;
        }

    }



    private void runVotingFlow () {
        logger.info("entered the voting flow");

        queue.add(new RestartVotingCommand(generateRequestIdentifier(), state.currentBallotSerialNumber));

        while (! wasShutDownCalled()) {
            try {
                ControllerCommand task = queue.take();
                handleSingleTask (task);
            }
            catch (InterruptedException e) {
                logger.warn ("Interrupted while reading from task queue " + e);
            }
        }
    }

    private void doShutDown () {
        logger.info("running shutDown");
    }

    private void handleSingleTask (ControllerCommand task) {
        if (task instanceof RestartVotingCommand) {
            doRestartVoting ();
            return;
        }
        if (task.getBallotSerialNumber() != state.currentBallotSerialNumber) {
            String errorMessage = "handleSingleTask: received a task too old. " +
                    task.getBallotSerialNumber() + " " + state.currentBallotSerialNumber;
            logger.warn (errorMessage);
            return;
        }

        if (task instanceof CastCommand) {
            doCast();
            return;
        }
        else if (task instanceof AuditCommand) {
            doAudit();
            return;
        }
        else if (task instanceof ChannelChoiceCommand) {
            doChooseChannel();
            return;
        }
        else if (task instanceof ChannelDeterminedCommand) {
            doSetChannelAndAskQuestions ((ChannelDeterminedCommand)task);
            return;
        }
        else if (task instanceof EncryptAndCommitBallotCommand) {
            doCommit ((EncryptAndCommitBallotCommand)task);
            return;
        }
        else {
            String errorMessage = "handleSingleTask: unknown type of ControllerCommand received: " +
                    task.getClass().getName();
            logger.error(errorMessage);
            throw new RuntimeException(errorMessage);
        }
    }

    /*
    Map<string, funcImpl> taskHandlers = new Map<>() {
CaskTask.getClass().getName() : doCast(),
ChannelChoiceCommand.getClass().getName() : doChooseChannel(),
}



     */

    private synchronized boolean wasShutDownCalled () {
        return shutDownHasBeenCalled;
    }

    private void doRestartVoting () {
        queue.clear();
        state.clearPlaintext();
        state.clearCiphertext();
        state.stateIdentifier = VBState.NEW_VOTER;
        state.currentBallotSerialNumber += 1;

        ui.startNewVoterSession(new NewVoterCallback(generateRequestIdentifier(), state.currentBallotSerialNumber , this.queue));
    }

    private void doChooseChannel () {
        if (state.stateIdentifier == VBState.NEW_VOTER) {
            logger.debug("doing chooseChannel");
            state.stateIdentifier = VBState.CHOOSE_CHANNEL;
            ui.chooseChannel(this.questionsForChoosingChannel,
                    new ChannelChoiceCallback(generateRequestIdentifier(), state.currentBallotSerialNumber, this.queue));
        }
        else {
            logger.warn("doChooseChannel: current state is " + state.stateIdentifier);
            // ignore this request
        }
    }

    private void doSetChannelAndAskQuestions (ChannelDeterminedCommand task) {
        if (state.stateIdentifier == VBState.CHOOSE_CHANNEL) {
            logger.debug("doing set channel and ask questions");
            state.stateIdentifier = VBState.ANSWER_QUESTIONS;
            BallotAnswer[] channelChoiceAnswers = task.channelChoiceAnswers;
            state.channelIdentifier = questionSelector.getChannelIdentifier(channelChoiceAnswers);
            state.channelSpecificQuestions = questionSelector.selectQuestionsForVoter(channelChoiceAnswers);
            state.stateIdentifier = VBState.ANSWER_QUESTIONS;
            ui.askVoterQuestions(state.channelSpecificQuestions,
                    new VotingCallback(generateRequestIdentifier(), state.currentBallotSerialNumber, this.queue));
        }
        else {
            logger.warn("doSetChannelAndAskQuestions: current state is " + state.stateIdentifier);
            // ignore this request
        }
    }


    private void doCommit (EncryptAndCommitBallotCommand task) {
        if (state.stateIdentifier == VBState.ANSWER_QUESTIONS) {
            logger.debug("doing commit");
            state.stateIdentifier = VBState.COMMITTING_TO_BALLOT;
            VotingBoothEncryptor.EncryptionAndSecrets encryptionAndSecrets = encryptor.encrypt(state.plaintextBallot);
            state.encryptedBallot = encryptionAndSecrets.getEncryptedBallot();
            state.secrets = encryptionAndSecrets.getSecrets();
            ui.notifyVoterToWaitForFinish(waitForCommitMessage,
                    new WaitForFinishCallback(generateRequestIdentifier(),
                            state.currentBallotSerialNumber,
                            this.queue));
            outputDevice.commitToBallot(state.plaintextBallot,
                    state.encryptedBallot,
                    new OutputDeviceCallback(generateRequestIdentifier(), state.currentBallotSerialNumber, this.queue));
        }
        else {
            logger.warn("doCommit: current state is " + state.stateIdentifier);
            // ignore this request
        }
    }


    private void doAudit () {
        if (state.stateIdentifier == VBState.CAST_OR_AUDIT) {
            logger.debug("doing audit");
            state.stateIdentifier = VBState.FINALIZING;
            outputDevice.audit(state.secrets,
                    new OutputDeviceCallback(generateRequestIdentifier(), state.currentBallotSerialNumber, this.queue));
            ui.notifyVoterToWaitForFinish(waitForAuditMessage,
                    new WaitForFinishCallback(generateRequestIdentifier(), state.currentBallotSerialNumber, this.queue));
        }
        else {
            logger.warn("doAudit: current state is " + state.stateIdentifier);
            // ignore this request
        }
    }

    private void doCast () {
        if (state.stateIdentifier == VBState.CAST_OR_AUDIT) {
            logger.debug("casting ballot");
            state.stateIdentifier = VBState.FINALIZING;
            outputDevice.castBallot(
                    new OutputDeviceCallback(generateRequestIdentifier(), state.currentBallotSerialNumber, this.queue));
            ui.notifyVoterToWaitForFinish(waitForCastMessage,
                    new WaitForFinishCallback(generateRequestIdentifier(), state.currentBallotSerialNumber, this.queue));
        }
        else {
            logger.warn("doCast: current state is " + state.stateIdentifier);
            // ignore this request
        }
    }





    private enum VBState {
        NEW_VOTER,
        CHOOSE_CHANNEL,
        ANSWER_QUESTIONS,
        COMMITTING_TO_BALLOT,
        CAST_OR_AUDIT,
        FINALIZING
    }


    private class ControllerState {
        public VotingBoothImpl.VBState stateIdentifier;
        public int channelIdentifier;
        public BallotQuestion[] channelSpecificQuestions;
        public PlaintextBallot plaintextBallot;
        public EncryptedBallot encryptedBallot;
        public BallotSecrets secrets;
        public int lastRequestIdentifier;
        public long currentBallotSerialNumber;

        public ControllerState () {
            plaintextBallot = null;
            encryptedBallot = null;
            secrets = null;
            lastRequestIdentifier = -1;
            channelIdentifier = -1;
            channelSpecificQuestions = null;
            currentBallotSerialNumber = 0;
        }


        public void clearPlaintext () {
            //TODO: Do we need safe erasure?
            plaintextBallot = null;
        }

        public void clearCiphertext () {
            //TODO: Do we need safe erasure?
            encryptedBallot = null;
            secrets = null;
        }
    }


    private int generateRequestIdentifier() {
        ++requestCounter;
        if (requestCounter >= MAX_REQUEST_IDENTIFIER) {
            requestCounter = 1;
        }
        return requestCounter;
    }


}

