package meerkat.voting.controller.selector;

import jdk.nashorn.internal.runtime.regexp.joni.exception.ValueException;
import meerkat.protobuf.Voting.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;

/**
 * Created by hai on 02/05/16.
 */
public class SimpleListCategoriesSelector extends QuestionSelector {
    protected final static Logger logger = LoggerFactory.getLogger(SimpleListCategoriesSelector.class);
    private boolean[] isSelected;

    public SimpleListCategoriesSelector(Object selectionData, BallotQuestion[] allBallotQuestions) {
        super(selectionData, allBallotQuestions);
        initializeAttributes (allBallotQuestions);
    }

    public SimpleListCategoriesSelector(Object selectionData, ArrayList<BallotQuestion> allBallotQuestions) {
        super(selectionData, null);
        BallotQuestion[] questionsArr = new BallotQuestion[allBallotQuestions.size()];
        for (int i = 0; i < questionsArr.length; ++i) {
            questionsArr[i] = allBallotQuestions.get(i);
        }
        initializeAttributes (questionsArr);
    }

    private void initializeAttributes (BallotQuestion[] allBallotQuestions) {
        assert selectionData instanceof SimpleCategoriesData;
        this.allBallotQuestions = allBallotQuestions;
        SimpleCategoriesData data = (SimpleCategoriesData) selectionData;
        int maxQuestionNumber = data.getMaxQuestionNumber();
        int length = allBallotQuestions.length;
        if (maxQuestionNumber >= length) {
            String errorMessage = "Selection data refers to question index " + maxQuestionNumber + " while we have only " + length + " questions totally";
            logger.error(errorMessage);
            throw new ValueException(errorMessage);
        }
        isSelected = new boolean[allBallotQuestions.length];
    }


    @Override
    public int getChannelIdentifier(BallotAnswer[] channelChoiceAnswers) {
        return 0;
    }

    @Override
    public BallotQuestion[] selectQuestionsForVoter(BallotAnswer[] channelChoiceAnswers) {
        SimpleCategoriesData data = (SimpleCategoriesData)selectionData;
        java.util.Arrays.fill(isSelected, false);
        markSelected(data.getSharedDefaults());
        markSelectionsOfVoterChannelChoice (channelChoiceAnswers);
        int[] questionIndices = extractSelectedIndices();
        BallotQuestion[] selectedQuestions = new BallotQuestion[questionIndices.length];
        for (int i = 0; i < questionIndices.length; ++i) {
            selectedQuestions[i] = allBallotQuestions[questionIndices[i]];
        }
        return selectedQuestions;
    }

    private void markSelectionsOfVoterChannelChoice(BallotAnswer[] channelChoiceAnswers) {
        SimpleCategoriesData data = (SimpleCategoriesData)selectionData;
        for (int q = 0; q < channelChoiceAnswers.length; ++q) {
            BallotAnswer ballotAnswer = channelChoiceAnswers[q];
            assertAnswerLengthIsOne(ballotAnswer, q);
            SingleChosenAnswer key = new SingleChosenAnswer(q, ballotAnswer.getAnswer(0));
            assertKeyInSelectionData(key);
            markSelected(data.getItem(key));
        }
    }

    private void assertAnswerLengthIsOne (BallotAnswer ballotAnswer, int questionNumber) {
        if (ballotAnswer.getAnswerCount() != 1) {
            String errorMessage = "SimpleListCategoriesSelector expects a single answer for every channel choice question\n";
            errorMessage += "Answer to question number " + (questionNumber+1) + " is";
            for (long i : ballotAnswer.getAnswerList()) {
                errorMessage += " " + i;
            }
            logger.error(errorMessage);
            throw new ValueException(errorMessage);
        }
    }

    private void assertKeyInSelectionData(SingleChosenAnswer key) {
        SimpleCategoriesData data = (SimpleCategoriesData)selectionData;
        int[] questionListToAdd = data.getItem(key);
        if (null == questionListToAdd) {
            String errorMessage = "SimpleListCategoriesSelector could not resolve answer number " + key.answerNumber + " for question number " + key.questionNumber;
            logger.error(errorMessage);
            throw new ValueException(errorMessage);
        }
    }

    private void markSelected (int[] questionIndices) {
        for (int i: questionIndices) {
            isSelected[i] = true;
        }
    }

    private int[] extractSelectedIndices() {
        int nSelected = 0;
        for (boolean b: isSelected) {
            if (b) {
                nSelected++;
            }
        }
        int[] res = new int[nSelected];
        int currIndex = 0;
        for (int questionNumber = 0; questionNumber < isSelected.length; ++questionNumber) {
            if (isSelected[questionNumber]) {
                res[currIndex] = questionNumber;
                currIndex++;
            }
        }
        return res;
    }
}
