package meerkat.voting.output;

import com.google.common.util.concurrent.FutureCallback;
import com.google.protobuf.ByteString;
import meerkat.protobuf.Crypto.*;
import meerkat.protobuf.Voting.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A toy OutputDevice class
 * outputs everything simply to the System console
 */
public class SystemConsoleOutputDevice implements BallotOutputDevice {

    private Logger logger;

    public SystemConsoleOutputDevice () {
        logger = LoggerFactory.getLogger(SystemConsoleOutputDevice.class);
        logger.info("A SystemConsoleOutputDevice is constructed");
    }

    /*
     * Returns the UTF8 decoding of byte-string data
     */
    private static String bytesToString(ByteString data) {
        return data.toStringUtf8();
    }

    @Override
    public void commitToBallot(PlaintextBallot plaintextBallot,
                               EncryptedBallot encryptedBallot,
                               FutureCallback<Void> callback) {
        logger.debug("entered method commitToBallot");
        long plaintextSerialNumber = plaintextBallot.getSerialNumber();
        System.out.println("Commitment of Ballot #" + plaintextSerialNumber + " (plaintext):");
        long encryptedSerialNumber = encryptedBallot.getSerialNumber();
        System.out.println("Commitment of Ballot #" + encryptedSerialNumber + " (ciphertext):");
        if (plaintextSerialNumber != encryptedSerialNumber) {
            logger.error("plaintext and encryption serial numbers do not match!! plaintext# = " +
                    plaintextSerialNumber + ", ciphertext# = " + encryptedSerialNumber);
        }
        ByteString encryptedData = encryptedBallot.getData().getData();
        System.out.println(bytesToString(encryptedData));
        callback.onSuccess(null);
    }

    @Override
    public void audit(BallotSecrets ballotSecrets, FutureCallback<Void> callback) {
        logger.debug("entered method audit");
        System.out.println("Auditing");
        printPlaintextBallot (ballotSecrets.getPlaintextBallot());
        printEncryptionRandomness (ballotSecrets.getEncryptionRandomness());
        printRandomnessGenerationProof (ballotSecrets.getProof());
        callback.onSuccess(null);
    }

    private void printPlaintextBallot (PlaintextBallot plaintextBallot) {
        long plaintextSerialNumber = plaintextBallot.getSerialNumber();
        System.out.println("Plaintext serial number = " + plaintextSerialNumber);

        System.out.println("Answers = ");
        for (BallotAnswer ballotAnswer : plaintextBallot.getAnswersList()) {
            String printableAnswer = "";
            for (long n : ballotAnswer.getAnswerList()) {
                printableAnswer += n + " ";
            }
            System.out.println(printableAnswer);
        }
    }

    private void printEncryptionRandomness (EncryptionRandomness encryptionRandomness) {
        System.out.println("Encryption Randomness = ");
        ByteString data = encryptionRandomness.getData();
        System.out.println(bytesToString(data));
    }

    private void printRandomnessGenerationProof (RandomnessGenerationProof proof) {
        System.out.println("Proof of randomness generation:");
        ByteString data = proof.getData();
        System.out.println(bytesToString(data));
    }


    @Override
    public void castBallot(FutureCallback<Void> callback) {
        logger.debug("entered method castBallot");
        System.out.println("Ballot finalized for casting!");
        callback.onSuccess(null);
    }

    @Override
    public void cancelBallot(FutureCallback<Void> callback) {
        logger.debug("entered method cancelBallot");
        System.out.println("Ballot cancelled!");
        callback.onSuccess(null);
    }
}
