package meerkat.voting.ui;

import com.google.common.util.concurrent.FutureCallback;
import com.google.protobuf.ByteString;
import meerkat.protobuf.Voting.*;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Date;
import java.util.StringTokenizer;
import java.util.Timer;
import java.util.concurrent.LinkedBlockingQueue;

import meerkat.voting.controller.callbacks.*;
import meerkat.voting.ui.uicommands.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import static java.lang.System.in;


public class SystemConsoleUI implements VotingBoothUI, Runnable {

    private BufferedReader bufferedReader;
    private LinkedBlockingQueue<UICommand> queue;
    private final Logger logger;
    private final int tickDurationInMillisec = 10;
    private Date startWaitingTime;


    public SystemConsoleUI() {
        logger = LoggerFactory.getLogger(SystemConsoleUI.class);
        logger.info("A VB UI console is constructed");
        queue = new LinkedBlockingQueue<>();
        bufferedReader = new BufferedReader(new InputStreamReader(in));

        startWaitingTime = null;
        Timer timer = new Timer();
        timer.scheduleAtFixedRate(new TickerTimerTask(queue), new Date(), tickDurationInMillisec);
    }


    public void run () {
        logger.info("entered the voting flow");
        while (true) {
            try {
                UICommand task = queue.take();
                handleSingleTask (task);
            }
            catch (InterruptedException e) {
                logger.warn ("Interrupted while reading from task queue " + e);
            }
        }
    }


    private void handleSingleTask (UICommand task) {
        if (!(task instanceof TickCommand)) {
            if (startWaitingTime != null) {
                stopWaiting();
            }
        }

        if (task instanceof StartSessionUICommand) {
            doShowWelcomeScreen((StartSessionUICommand)task);
            return;
        }
        else if (task instanceof ChannelChoiceUICommand) {
            doAskChannelChoiceQuestions((ChannelChoiceUICommand)task);
            return;
        }
        else if (task instanceof RaceVotingUICommand) {
            doAskVotingQuestions((RaceVotingUICommand)task);
            return;
        }
        else if (task instanceof CastOrAuditUICommand) {
            doCastOrAudit ((CastOrAuditUICommand)task);
            return;
        }
        else if (task instanceof HaltCommand) {
            doHalt((HaltCommand)task);
            return;
        }
        else if (task instanceof WaitForFinishCommand) {
            doWaitForFinish((WaitForFinishCommand)task);
            return;
        }
        else if (task instanceof TickCommand) {
            doTick ();
            return;
        }
        else {
            String errorMessage = "handleSingleTask: unknown type of UICommand received: " +
                    task.getClass().getName();
            logger.error(errorMessage);
            throw new RuntimeException(errorMessage);
        }
    }


    @Override
    public void startNewVoterSession(FutureCallback<Void> callback) {
        logger.debug("UI interface call to startNewVoterSession");
        queue.add(new StartSessionUICommand((ControllerCallback)callback));
    }

    private void doShowWelcomeScreen(StartSessionUICommand task) {
        logger.debug("UI entered doShowWelcomeScreen");
        System.out.println("Welcome, new voter!");
        waitForEnter(null);
        ControllerCallback callback = task.getCallback();
        assert (callback instanceof NewVoterCallback);
        ((NewVoterCallback)callback).onSuccess(null);
    }

    private void stopWaiting () {
        System.out.println ();
        startWaitingTime = null;

    }
    private void waitForEnter(String message) {
        if (message != null) {
            System.out.println(message);
        }
        System.out.println("\nPress ENTER to proceed.\n");
        String t = null;
        while (t == null) {
            try {
                t = readInputLine();
            }
            catch (IOException e) {
                String errorMessage = "waitForEnter: threw an IOException: " + e;
                logger.error(errorMessage);
                System.err.println(errorMessage);
            }
        }
    }

    @Override
    public void chooseChannel(BallotQuestion[] questions, FutureCallback<BallotAnswer[]> callback) {
        logger.debug("UI interface call to chooseChannel");
        ChannelChoiceUICommand task = new ChannelChoiceUICommand(questions, (ControllerCallback)callback);
        queue.add(task);
    }

    private void doAskChannelChoiceQuestions (ChannelChoiceUICommand task) {
        logger.debug("UI: doAskChannelChoiceQuestions");
        System.out.println("Showing questions for choosing channel:\n");
        try {
            BallotAnswer[] answers = askVoterForAnswers(task.getQuestions());
            task.getCallback().onSuccess(answers);
        }
        catch (IOException e) {
            String errorMessage = "Channel choice failed due to IOException: " + e;
            logger.error (errorMessage);
            System.err.println(errorMessage);
            task.getCallback().onFailure(e);
        }
    }

    @Override
    public void askVoterQuestions(BallotQuestion[] questions, FutureCallback<BallotAnswer[]> callback) {
        logger.debug("UI interface call to chooseChannel");
        queue.add(new RaceVotingUICommand(questions, (ControllerCallback)callback));
    }

    private void doAskVotingQuestions (RaceVotingUICommand task) {
        logger.debug("UI: doAskVotingQuestions");
        System.out.println("Showing questions for race voting:\n");
        try {
            BallotAnswer[] answers = askVoterForAnswers(task.getQuestions());
            task.getCallback().onSuccess(answers);
        }
        catch (IOException e) {
            String errorMessage = "Asking voting questions failed due to IOException: " + e;
            logger.error (errorMessage);
            System.err.println(errorMessage);
            task.getCallback().onFailure(e);
        }
    }

    @Override
    public void castOrAudit(FutureCallback<FinalizeBallotChoice> callback) {
        logger.debug("UI interface call to castOrAudit");
        queue.add(new CastOrAuditUICommand((ControllerCallback)callback));
    }

    private void doCastOrAudit(CastOrAuditUICommand task) {
        logger.debug("UI entered doCastOrAudit");
        System.out.println ("Finalizing your vote. Do you wish to (C)ast or (A)udit?");

        FinalizeBallotChoice fChoice;

        try {
            String s = readInputLine();
            if (s.equals("cast") || s.equals("c")) {
                fChoice = FinalizeBallotChoice.CAST;
            }
            else if (s.equals("audit") || s.equals("a")) {
                fChoice = FinalizeBallotChoice.AUDIT;
            }
            else {
                throw new IllegalArgumentException("UI could not understand the answer for cast/audit question '" + s + "'");
            }
            ControllerCallback callback = task.getCallback();
            assert (callback instanceof CastOrAuditCallback);
            ((CastOrAuditCallback)callback).onSuccess(fChoice);
        }
        catch (IllegalArgumentException|IOException e) {
            String errorMessage = "doCastOrAudit: some error with reading input from console. details: " + e;
            logger.error(errorMessage);
            task.getCallback().onFailure(e);
        }
    }


    @Override
    public void notifyVoterToWaitForFinish(UIElement message, FutureCallback<Boolean> callback) {
        logger.debug("UI interface call to notifyVoterToWaitForFinish");
        queue.add(new WaitForFinishCommand(message, (ControllerCallback)callback));
    }

    public void doWaitForFinish (WaitForFinishCommand task) {
        logger.debug("UI entered doWaitForFinish");

        startWaitingTime = new Date();

        UIElement message = task.getMessage();
        String messageString;
        if (message.getType() != UIElementDataType.TEXT) {
            messageString = "Default message: encountered an error. System halting";
        } else {
            messageString = bytesToString(message.getData());
        }
        System.out.println(messageString);
        System.out.print ("Waiting : .");
    }

    @Override
    public void showErrorMessageAndHalt(UIElement errorMessage, FutureCallback<Boolean> callback) {
        logger.debug("UI interface call to showErrorMessageAndHalt");
        queue.add(new HaltCommand(errorMessage, (ControllerCallback)callback));
    }

    private void doHalt (HaltCommand task) {
        logger.debug("UI entered doHalt");

        UIElement errorMessage = task.getErrorMessage();
        String errorMessageString;
        if (errorMessage.getType() != UIElementDataType.TEXT) {
            errorMessageString = "Default message: encountered an error. System halting";
        } else {
            errorMessageString = bytesToString(errorMessage.getData());
        }
        System.out.println(errorMessageString);

        while (true) {
            try {
                wait();
            } catch (InterruptedException e) {
                logger.error("UI halt has been interrupted. Details: " + e);
                ControllerCallback callback = task.getCallback();
                assert callback instanceof HaltCallback;
                ((HaltCallback) callback).onFailure(e);
            }
        }
    }


    @Override
    public void showErrorMessageWithButtons(UIElement errorMessage, UIElement[] buttonLabels, FutureCallback<Integer> callback) {
        throw new UnsupportedOperationException();
    }

    private void doTick () {
        if (startWaitingTime != null) {
            System.out.print (".");  // still waiting
        }
    }

    private String readInputLine() throws IOException{
        String s;
        try {
            s = this.bufferedReader.readLine();
            if (null == s) {
                throw new IOException();
            }
        } catch (IOException e) {
            String errorMessage = "readInputLine: some error with reading input from console. details: " + e;
            logger.error(errorMessage);
            throw new IOException(e);
        }
        return s;
    }


    private void assertQuestionsAreValid (BallotQuestion[] questions) {
        for (int index = 0; index < questions.length; ++index) {
            BallotQuestion question = questions[index];
            if (question.getIsMandatory()) {
                String errorMessage = "askVoterQuestions: question number " + index + " is marked as mandatory";
                logger.error(errorMessage);
                throw new UnsupportedOperationException(errorMessage);
            }
            if (!isQuestionOnlyText(question)) {
                String errorMessage = "askVoterQuestions: question number " + index + " is not only text";
                logger.error(errorMessage);
                throw new UnsupportedOperationException(errorMessage);
            }
        }
    }

    private BallotAnswer[] askVoterForAnswers(BallotQuestion[] questions) throws IOException {

        assertQuestionsAreValid (questions);

        BallotAnswer answers[] = new BallotAnswer[questions.length];
        int index = 0;
        while (index < questions.length) {
            BallotQuestion question = questions[index];
            System.out.println("Question number " + index);
            showQuestionOnScreen(question);

            System.out.println("UI screen: Enter your answer. You can also type '(b)ack' or '(c)ancel' or '(s)kip");
            String s = readInputLine();

            if ((s.equals("cancel") || s.equals("c")) || (index == 0 && (s.equals("back") || s.equals("b")))) {
                return null;
            }
            else if (s.equals("back") || s.equals("b")) {
                --index;
            }
            else if (s.equals("skip") || s.equals("s")) {
                answers[index] = translateStringAnswerToProtoBufMessageAnswer("");
                ++index;
            }
            else {
                answers[index] = translateStringAnswerToProtoBufMessageAnswer(s);
                ++index;
            }
        }
        return answers;
    }

    //show question on the screen for the voter
    private void showQuestionOnScreen(BallotQuestion question) {
        if (!isQuestionOnlyText(question)) {
            System.err.println("debug: an element in question is not of TEXT type");
            throw new UnsupportedOperationException();
        }

        System.out.println("Question text: " + bytesToString(question.getQuestion().getData()));
        System.out.println("Description: " + bytesToString(question.getDescription().getData()));
        int answerIndex = 0;
        for (UIElement answer : question.getAnswerList()) {
            ++answerIndex;
            System.out.println("Answer " + answerIndex + ": " + bytesToString(answer.getData()));
        }
    }

    private boolean isQuestionOnlyText (BallotQuestion question) {
        boolean isText = true;
        if (question.getQuestion().getType() != UIElementDataType.TEXT
                || question.getDescription().getType() != UIElementDataType.TEXT) {
            isText = false;
        }
        for (UIElement answer : question.getAnswerList()) {
            if (answer.getType() != UIElementDataType.TEXT) {
                isText = false;
            }
        }
        return isText;
    }


    private BallotAnswer translateStringAnswerToProtoBufMessageAnswer(String s) {
        BallotAnswer.Builder bab = BallotAnswer.newBuilder();
        StringTokenizer st = new StringTokenizer(s);
        while (st.hasMoreTokens()) {
            bab.addAnswer(Integer.parseInt(st.nextToken()));
        }
        return bab.build();
    }

    /*
     * Returns the UTF8 decoding of byte-string data
     */
    private static String bytesToString(ByteString data) {
        return data.toStringUtf8();
    }

}
